/***************************************************************************
 * Copyright 2012 by
 *  + Christian-Albrechts-University of Kiel
 *    + Department of Computer Science
 *      + Software Engineering Group 
 *  and others.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ***************************************************************************/

package kieker.webgui.beans.session;

import java.io.IOException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;

import javax.faces.application.FacesMessage;
import javax.faces.application.FacesMessage.Severity;
import javax.faces.bean.ManagedBean;
import javax.faces.bean.SessionScoped;
import javax.faces.context.FacesContext;

import kieker.analysis.display.annotation.Display;
import kieker.analysis.model.analysisMetaModel.MIAnalysisMetaModelFactory;
import kieker.analysis.model.analysisMetaModel.MIDisplay;
import kieker.analysis.model.analysisMetaModel.MIPlugin;
import kieker.analysis.model.analysisMetaModel.MIProject;
import kieker.analysis.model.analysisMetaModel.MIView;
import kieker.analysis.model.analysisMetaModel.impl.MAnalysisMetaModelFactory;
import kieker.analysis.plugin.AbstractPlugin;
import kieker.webgui.common.FSManager;
import kieker.webgui.common.Pair;
import kieker.webgui.common.exception.NewerProjectException;

import org.primefaces.event.TabChangeEvent;

/**
 * This bean contains the project of the current (session) user for the analysis view work space.
 * 
 * @author Nils Christian Ehmke
 * @version 1.0
 */
// TODO Display Semantic
@ManagedBean
@SessionScoped
public class CurrentAnalysisViewWorkSpaceProjectBean {

	/**
	 * This is the factory which will be used to create new components for the project.
	 */
	private final MIAnalysisMetaModelFactory factory = new MAnalysisMetaModelFactory();
	/**
	 * This is the page used for the redirecting during setting the current project of the user.
	 */
	private static final String PAGE_ANALYSIS_VIEW_WORK_SPACE = "AnalysisViewWorkSpace.xhtml";
	/**
	 * This is the page used for the redirecting during closing the work space.
	 */
	private static final String PAGE_PROJECT_OVERVIEW = "ProjectOverview.xhtml";
	/**
	 * This is the name of the stored project. It can be used as an identifier within the FS-Manager
	 */
	private String projectName;
	/**
	 * This is the actual model instance. It is the in-memory-model of the current (session) user.
	 */
	private MIProject project;
	/**
	 * This is the time stamp of the moment, the project was loaded or last saved. It can be used to check whether the project has been modified in the meanwhile.
	 */
	private long timeStamp;
	/**
	 * This is the corresponding class loader to the project. It contains always the libraries within the lib-folder of the project.
	 */
	private ClassLoader classLoader;
	/**
	 * This is the currently selected view.
	 */
	private MIView activeView;

	/**
	 * Creates a new instance of this class.
	 */
	public CurrentAnalysisViewWorkSpaceProjectBean() {
		// No code necessary
	}

	/**
	 * This method delivers the project stored in this bean.
	 * 
	 * @return The project for this user.
	 */
	public MIProject getProject() {
		synchronized (this) {
			return this.project;
		}
	}

	/**
	 * This method sets the project stored within this bean and returns the new page for the navigation.
	 * 
	 * @param name
	 *            The name of the project.
	 * @return The name of the page for the analysis view work space.
	 */
	public String setProject(final MIProject newProject, final String newName) {
		synchronized (this) {
			// Remember the given parameters
			this.project = newProject;
			this.projectName = newName;

			if (this.project != null) {
				// Remember the current time! This is important for the later comparison of the time stamps.
				this.resetTimeStamp();
				this.reloadClassLoader();
			}
		}

		return CurrentAnalysisViewWorkSpaceProjectBean.PAGE_ANALYSIS_VIEW_WORK_SPACE;
	}

	/**
	 * This method delivers the project name stored in this bean.
	 * 
	 * @return The project name for this user.
	 */
	public String getProjectName() {
		synchronized (this) {
			return this.projectName;
		}
	}

	/**
	 * This method clears the bean. In other words: The stored project is set to null and it will return the page of the project overview for navigation purposes.
	 * 
	 * @return The name of the page of the project overview.
	 */
	public String clearProject() {
		synchronized (this) {
			this.projectName = null; // NOPMD
			this.project = null; // NOPMD
			this.classLoader = null; // NOPMD
			this.activeView = null; // NOPMD
		}

		return CurrentAnalysisViewWorkSpaceProjectBean.PAGE_PROJECT_OVERVIEW;
	}

	/**
	 * This method tries to save the current project and informs the user about success or fail.
	 * 
	 * @param overwriteNewerProject
	 *            This flag determines whether a newer project should be overwritten.
	 */
	public void saveProject(final boolean overwriteNewerProject) {
		synchronized (this) {
			try {
				FSManager.getInstance().saveProject(this.projectName, this.project, this.timeStamp, overwriteNewerProject);
				CurrentAnalysisViewWorkSpaceProjectBean.showMessage(FacesMessage.SEVERITY_INFO, "Project saved.");
				// Update the time stamp!
				this.resetTimeStamp();
			} catch (final IOException ex) {
				CurrentAnalysisViewWorkSpaceProjectBean.showMessage(FacesMessage.SEVERITY_ERROR, "An error occured while saving the projct.");
			} catch (final NewerProjectException ex) {
				CurrentAnalysisViewWorkSpaceProjectBean.showMessage(FacesMessage.SEVERITY_WARN, "The project has been modified externally in the meanwhile.");
			}
		}
	}

	/**
	 * This method shows the current user a message by using the growl-component of PrimeFaces.
	 * 
	 * @param severity
	 *            The severity of the message.
	 * @param msg
	 *            The message itself.
	 */
	private static void showMessage(final Severity severity, final String msg) {
		FacesContext.getCurrentInstance().addMessage(null, new FacesMessage(severity, "", msg));
	}

	/**
	 * This method sets the time stamp to the current system time.
	 */
	public void resetTimeStamp() {
		synchronized (this) {
			this.timeStamp = System.currentTimeMillis();
		}
	}

	/**
	 * This method delivers the available displays of the given plugin.
	 * 
	 * @param clazz
	 *            The plugin whose displays should be delivered.
	 * @return The available displays.
	 */
	public List<Display> getDisplays(final Class<AbstractPlugin> clazz) {
		final List<Display> result = new ArrayList<Display>();

		synchronized (this) {

			if (clazz != null) {
				// Run through all available methods and find the annotated ones
				final Method[] methods = clazz.getMethods();
				for (final Method method : methods) {
					final Display displayAnnot = method.getAnnotation(Display.class);
					if (displayAnnot != null) {
						result.add(displayAnnot);
					}
				}
			}
		}

		return result;
	}

	/**
	 * This method delivers a list of pairs containing the available plugins and their classes.
	 * 
	 * @return A list with the plugins.
	 */
	public List<Pair<MIPlugin, Class<AbstractPlugin>>> getPlugins() {
		final List<Pair<MIPlugin, Class<AbstractPlugin>>> result = new ArrayList<Pair<MIPlugin, Class<AbstractPlugin>>>();

		synchronized (this) {
			for (final MIPlugin plugin : this.project.getPlugins()) {
				try {
					result.add(new Pair<MIPlugin, Class<AbstractPlugin>>(plugin, (Class<AbstractPlugin>) this.classLoader.loadClass(plugin.getClassname())));
				} catch (final ClassNotFoundException ex) {
					// TODO Catch
				}
			}
		}

		return result;
	}

	/**
	 * This method adds a new view to the project.
	 * 
	 * @param viewName
	 *            The name of the new view.
	 */
	public void addView(final String viewName) {
		synchronized (this) {
			final MIView view = this.factory.createView();
			view.setName(viewName);
			this.project.getViews().add(view);
		}
	}

	/**
	 * Delivers the currently selected view.
	 * 
	 * @return The currently active view.
	 */
	public MIView getActiveView() {
		synchronized (this) {
			return this.activeView;
		}
	}

	/**
	 * This method sets the active view to a new value.
	 * 
	 * @param view
	 *            The new active view.
	 */
	public void setActiveView(final MIView view) {
		synchronized (this) {
			this.activeView = view;
		}
	}

	public void addDisplayToView(final MIDisplay display) {
		synchronized (this) {
			if (this.activeView != null) {
				this.activeView.getDisplays().add(display);
			}
		}
	}

	public void onChange(final TabChangeEvent event) {
		if (event.getData() instanceof MIView) {
			this.setActiveView((MIView) event.getData());
		}
	}

	/**
	 * This method reloads the class loader. In other words: The class loader will always be able to load classes from the jar-files within the lib-folder of the
	 * project.
	 */
	private void reloadClassLoader() {
		synchronized (this) {
			this.classLoader = FSManager.getInstance().getClassLoader(this.projectName);
		}
	}
}
